import 'dart:convert';
import 'package:flutter/cupertino.dart';

typedef ClearFunction = void Function();

class KeyChild {
  final String key;
  final String subKey;

  KeyChild(this.key, this.subKey);
}

enum ControlType { DATA, VALIDATION, VALIDATION_FUNCTION, CLEAR, FOCUS_NODE }

class YekongaForm {
  Map<String, dynamic> data = {};
  Map<String, bool> validationRequirements = {};
  Map<String, Function(bool)> validationFunctions = {};
  Map<String, FocusNode> focusNodes = {};
  Map<String, Function> clearFunctions = {};

  dynamic parentNextFocusNode;

  YekongaForm({Map<String, dynamic> values}) {
    if (data != null) {
      this.fromMap(values);
    }
  }

  void setValue(String key, dynamic value) {
    try {
      KeyChild keyChild = extractKeys(key);

      if (keyChild.subKey != null) {
        this._setObjectValue(keyChild.key, keyChild.subKey, value);
      } else if (value is Map<dynamic, dynamic>) {
        this._setValue(keyChild.key, YekongaForm(values: value));
      } else {
        this._setValue(keyChild.key, value);
      }
    } catch (e) {
      print(e);
    }
  }

  void _setValue(String key, dynamic value) {
    try {
      if (data.containsKey(key)) {
        data[key] = value;
      } else {
        data.putIfAbsent(key, () => value);
      }
    } catch (e) {
      print("Set normal form value fail");
    }
  }

  void _setObjectValue(String key, String subKey, dynamic value) {
    try {
      if (data.containsKey(key) && data[key] is YekongaForm) {
        data[key].setValue(subKey, value);
      } else {
        data.putIfAbsent(key, () {
          YekongaForm subData = YekongaForm();
          subData.setValue(subKey, value);
          return subData;
        });
      }
    } catch (e) {
      print("Set object form value fails");
    }
  }

  getValue(String key) {
    KeyChild keyChild = extractKeys(key);

    if (data.containsKey(keyChild.key)) {
      if (data[key] is YekongaForm) {
        if (keyChild.subKey != null) {
          return data[key].getValue(keyChild.subKey);
        }
      } else {
        return data[keyChild.key];
      }
    }

    return null;
  }

  KeyChild extractKeys(String key) {
    if (key != null && key.contains('.')) {
      List<String> keys = key.split('.');
      String _key = keys.removeAt(0);
      String _subKey = keys.join('.');

      return KeyChild(_key, _subKey);
    }

    return KeyChild(key, null);
  }

  bool containsKey(String key) {
    return data.containsKey(key);
  }

  void fromMap(Map<String, dynamic> _data) {
    if (_data != null) {
      _data.forEach((key, value) {
        setValue(key, value);
      });
    }
  }

  Map<String, dynamic> toMap() {
    Map<String, dynamic> _data = {};

    this.data.forEach((key, value) {
      if (value is YekongaForm) {
        _data[key] = value.toMap();
      } else {
        _data[key] = value;
      }
    });

    return _data;
  }

  String toJson() {
    return this.toString();
  }

  @override
  String toString() {
    return json.encode(this.toMap());
  }

  void setValidationFunction(String key, Function(bool) value) {
    _setGeneralFunction(
        validationFunctions, key, value, ControlType.VALIDATION_FUNCTION);
  }

  void setValidationRequirement(String key, bool required) {
    _setGeneralFunction(
        validationRequirements, key, required, ControlType.VALIDATION);
  }

  void setClearFunction(String key, Function action) {
    _setGeneralFunction(clearFunctions, key, action, ControlType.CLEAR);
  }

  void setFocusNode(String key, FocusNode node) {
    _setGeneralFunction(focusNodes, key, node, ControlType.FOCUS_NODE);
  }

  void _setGeneralFunction(Map<String, dynamic> targetList, String key,
      dynamic value, ControlType type) {
    KeyChild keyChild = extractKeys(key);

    try {
      if (keyChild.subKey != null) {
        if (!data.containsKey(keyChild.key) ||
            (data.containsKey(keyChild.key) &&
                !(data[keyChild.key] is YekongaForm))) {
          data.putIfAbsent(keyChild.key, () => YekongaForm());
        }

        if (type == ControlType.VALIDATION) {
          data[keyChild.key].setValidationRequirement(keyChild.subKey, value);
        } else if (type == ControlType.VALIDATION_FUNCTION) {
          data[keyChild.key].setValidationFunction(keyChild.subKey, value);
        } else if (type == ControlType.CLEAR) {
          data[keyChild.key].setClearFunction(keyChild.subKey, value);
        } else if (type == ControlType.FOCUS_NODE) {
          data[keyChild.key].setFocusNode(keyChild.subKey, value);
        }
      } else {
        targetList[keyChild.key] = value;
      }
    } catch (e) {
      print("Set ${type.toString()} function fails");
      print(e);
    }
  }

  FocusNode getNextFocus(String key) {
    int pos = 0;
    String nextKey;
    List<String> list = this.data.keys.toList();

    if (key == null) {
      nextKey = list.elementAt(pos);
    } else if (this.data.containsKey(key)) {
      pos = list.indexOf(key);
      if (pos < list.length - 1) {
        nextKey = list.elementAt(pos + 1);
      }
    }

    for (var i = 0; i < (list.length - 1); i++) {
      if (i > pos) {
        String k1 = list.elementAt(i);
        String k2 = list.elementAt(i + 1);

        if (this.data[k1] is YekongaForm) {
          this.data[k1].parentNextFocusNode = this.data[k2].getNextFocus(null);
        }
      }
    }

    if (nextKey != null) {
      if (this.data[nextKey] is YekongaForm) {
        return this.data[nextKey].getNextFocus(null);
      } else if (this.focusNodes.containsKey(nextKey)) {
        return this.focusNodes[nextKey];
      }
    }

    if (this.parentNextFocusNode != null) {
      return this.parentNextFocusNode;
    }

    return null;
  }

  void clearAll() {
    List<String> keys = this.data.keys.toList();
    for (var key in keys) {
      if (this.data[key] is YekongaForm) {
        this.data[key].clearAll();
      } else {
        this.data[key] = null;
        if (this.clearFunctions.containsKey(key)) {
          this.clearFunctions[key]();
        }
      }
    }
  }

  bool get isValid => (() {
        bool status = true;
        this.data.forEach((key, value) {
          bool _status = true;

          if (this.validationRequirements.containsKey(key) &&
              this.validationRequirements[key]) {
            if (this.data[key] == null || this.data[key] == "") {
              status = false;
              _status = false;
            } else if (this.data[key] is YekongaForm) {
              status = this.data[key].isValid;
            }
          }

          if (this.validationFunctions.containsKey(key)) {
            this.validationFunctions[key](_status);
          }
        });

        return status;
      }());
}
